from r3frame.globs import pg
from r3frame.atom import Atom

# ------------------------------------------------------------ #
class Process(Atom):
    def __init__(self, pid: int, app):
        super().__init__(pid, 0)
        self.app = app
        self._remove = lambda: self.app.processes.remove(self)

    def callback(self) -> bool: return False
    def fallback(self) -> None: self._remove()
# ------------------------------------------------------------ #



from r3frame.globs import os, re, pg
from r3frame.app.resource.base import Resource
# from r3frame.app.resource.sound import Sound
from r3frame.app.resource.image import Image
from r3frame.app.resource.animation import Animation

# ------------------------------------------------------------ #
class ResourceManager:
    def __init__(self) -> None:
        self.font:dict = {}
        self.sound:dict = {}
        self.image:dict[str, Image] = {}
        self.animation:dict[str, Animation] = {}

        self.create_surface = lambda size, color: pg.Surface(size)
        self.create_rect = lambda location, size: pg.Rect(location, size)

        self.fill_surface = lambda surface, color: surface.fill(color)
        self.flip_surface = lambda surface, x, y: pg.transform.flip(surface, x, y)
        self.scale_surface = lambda surface, scale: pg.transform.scale(surface, scale)
        self.rotate_surface = lambda surface, angle: pg.transform.rotate(surface, angle)

        self.blit_rect = lambda rect, color, width: self.draw_rect(rect.size, rect.topleft, color, width)
        self.draw_line = lambda surface, start, end, color, width: pg.draw.line(surface, color, start, end, width=width)
        self.draw_rect = lambda surface, size, location, color, width: pg.draw.rect(surface, color, pg.Rect(location, size), width=width)
        self.draw_circle = lambda surface, center, radius, color, width: pg.draw.circle(surface, color, [*map(int, center)], radius, width)

    def flip_image(self, key: str, x: bool, y: bool) -> None:
        try:
            image = self.image[key]
            if isinstance(image, pg.Surface):
                self.image[key] = self.flip_surface(image, x, y)
            elif isinstance(image, list):
                self.image[key] = [self.flip_surface(i, x, y) for i in image]
        except (KeyError) as err: print(err)

    def get_image(self, key:str) -> Image|None:
        return self.image.get(key, None)
    
    def set_image(self, key:str, image:pg.Surface) -> None:
        try:
            self.image[key].data = image
        except (KeyError) as err: print(err)

    def load_image(self, key:str, path:str, scale:list=None, colorKey:list=None) -> Image:
        try:
            image:pg.Surface = pg.image.load(path).convert_alpha()
            image.set_colorkey(colorKey)
            if scale: image = self.scale_surface(image, scale)
            
            rid = len(self.image.values())
            self.image[key] = Image(rid, path, image)
            return self.image[key]
        except (FileNotFoundError) as err: print(err)
    
    def load_animation(
            self,
            key: str, sheet_path: str,
            frame_size: list[int], scale: list=None, colorKey: list=None,
            loop: bool=True, frame_duration: int= 5, frame_offset: list[int]=[0, 0]
        ) -> Animation:
        try:
            frames = self.load_image_sheet(key, sheet_path, frame_size, scale, colorKey)
            rid = len(self.animation.values())
            self.animation[key] = Animation(rid, frames, loop, frame_duration, frame_offset)
            return self.animation[key]
        except FileNotFoundError as e: return None

    def load_image_dir(self, key:str, path:str, scale:list=None, colorKey:list=None) -> list:
        try:
            rid = len(self.image.values())
            images:list[Image] = []
            for _, __, image in os.walk(path):
                sorted_images = sorted(image, key=lambda string_: [int(s) if s.isdigit() else s for s in re.split(r'(\d+)', string_)])
                for image in sorted_images:
                    full_path = path + '/' + image
                    image_surface = self.load_image(full_path, scale, colorKey)
                    if self.image_visible(image_surface):
                        img = Image(rid, path, image_surface)
                        images.append(img)
                        rid += 1
            self.image[key] = images
            return self.image[key]
        except (FileNotFoundError) as err: ...
    
    def load_image_sheet(self, key: str, path: str, frame_size: list[int], scale: list=None, colorKey: list=None) -> list:
        try:
            sheet: Image = self.load_image(key, path)
            frame_x = int(sheet.size[0] / frame_size[0])
            frame_y = int(sheet.size[1] / frame_size[1])
            
            rid = len(self.image.values())
            frames: list[Image] = []
            for row in range(frame_y):
                for col in range(frame_x):
                    x = col * frame_size[0]
                    y = row * frame_size[1]
                    frame = Image(rid, path, pg.Surface(frame_size, pg.SRCALPHA).convert_alpha())
                    frame.data.set_colorkey(colorKey)
                    frame.data.blit(sheet.data, (0,0), pg.Rect((x, y), frame_size))   # blit the sheet at the desired coords (texture mapping)
                    if scale: frame.set_scale(scale)
                    if self.image_visible(frame):
                        frames.append(frame)
                        rid += 1
            self.image[key] = frames
            return self.image[key]
        except FileNotFoundError as e: return None

    def image_visible(self, image: Image, threshold: int=1) -> bool:
        result = False
        pixels, noPixels = 0, 0
        for y in range(image.size[1]):
            for x in range(image.size[0]):
                pixel = image.data.get_at([x, y])
                if pixel[3] == 0:
                    noPixels += 1
                pixels += 1
        if pixels-noPixels >= threshold:
            result = True
        return result
# ------------------------------------------------------------ #



from r3frame.globs import pg
from r3frame.util import abs_path
from r3frame.app.ui.base import Interface
from r3frame.game import Tilemap, StaticPartition

# ------------------------------------------------------------ #
class Scene:
    def __init__(
            self, app, name: str,
            world_cellsize: int=32,
            world_size: list[int]=[200, 200],
            display_size: list[int]=[1600, 1200],
            interface_size: list[int]=[100, 100],
            display_color: list[int]=[203, 219, 252],
            interface_location: list[int]=[0, 0],
            font_path: str=abs_path("assets/fonts/megamax.ttf"),
    ) -> None:
        self.app = app
        self.name: str = name
        self.display_size: list[int] = display_size
        self.display_color: list[int] = display_color
        self.display: pg.Surface = pg.Surface(display_size)
        self.interface: Interface = Interface(
            name="scene interface",
            window=app.window,
            size=interface_size,
            location=interface_location,
            title_color=[255, 255, 255],
            font_path=font_path,
            text_color=[0, 0, 0],
            text_size=24,
        )


        self.world_size: list[int] = world_size
        self.world_cellsize: int = world_cellsize
        self.tilemap: Tilemap = Tilemap(app, world_size, world_cellsize)
        self.partition: StaticPartition = StaticPartition(world_size, world_cellsize)
        
        self.configure()

    def configure(self) -> None: ...
    def handle_events(self) -> None: raise NotImplementedError
    def handle_update(self) -> None: raise NotImplementedError
    def handle_render(self) -> None: raise NotImplementedError
# ------------------------------------------------------------ #



from r3frame.globs import pg

# ------------------------------------------------------------ #
class Window:
    def __init__(self, size: list[int], display_size: list[int], color: list[int]=[25, 25, 25]) -> None:
        self.icon = None
        self.title = "HFWindow"
        self.size = size
        self.color = color
        self.clip_range = [1, 1]
        self.display_size = display_size
        self.window = pg.display.set_mode(size)
        
        self.blit_rect = lambda rect, color, width: self.draw_rect(rect.size, rect.topleft, color, width)
        self.draw_line = lambda start, end, color, width: pg.draw.line(self.display, color, start, end, width=width)
        self.draw_rect = lambda size, location, color, width: pg.draw.rect(self.display, color, pg.Rect(location, size), width=width)
        self.draw_circle = lambda center, radius, color, width: pg.draw.circle(self.display, color, [*map(int, center)], radius, width)
        self.configure(display_size)

    def set_title(self, title: str) -> None:
        self.title = title
        self.configure()

    def set_icon(self, icon: pg.Surface) -> None:
        self.icon = icon
        self.configure()

    def configure(self, display_size: list[int]) -> None:
        self.display_size = display_size
        self.display = pg.Surface(display_size)
        if isinstance(self.title, str): pg.display.set_caption(self.title)
        if isinstance(self.icon, pg.Surface): pg.display.set_icon(self.icon)

    def clear(self) -> None:
        self.display.fill(self.color)
        self.window.fill(self.color)

    def blit(self, surface: pg.Surface, location: list[int], offset: list[int]=[0, 0]) -> None:
        # display-culling
        if ((location[0] + surface.size[0]) - self.clip_range[0] < 0 or location[0] + self.clip_range[0] > self.display_size[0]) \
        or ((location[1] + surface.size[1]) - self.clip_range[1] < 0 or location[1] + self.clip_range[1] > self.display_size[1]):
            return
        self.display.blit(surface, [location[0] - offset[0], location[1] - offset[1]])

    def update(self) -> None:
        pg.display.flip()
# ------------------------------------------------------------ #



from r3frame.atom import Atom
from r3frame.globs import os, pg
from r3frame.app.input import Mouse
from r3frame.app.window import Window
from r3frame.util import point_inside
from r3frame.app.ui.element import ELEMENT_STATE, Element

# ------------------------------------------------------------ #
class Interface(Atom):
    def __init__(
            self, name: str, window: Window,
            size: list[int], location: list[float],
            font_path:str, title_color:list[int]=[255, 255, 255],
            text_color:list[int]=[255, 255, 255], text_size: int=18
        ) -> None:
        super().__init__(0, 0)
        pg.font.init()
        self.name = name
        self.window = window
        self.size: list[int] = size
        self.location: list[float] = location

        self.elements: dict[str, Element] = {}
        
        self.font_path = font_path
        self.text_size: int = text_size
        self.text_fields: dict[str, str] = {}
        self.text_color: list[int] = text_color
        self.title_color: list[int] = title_color
        self.font: pg.Font = pg.Font(font_path, text_size)

        self.show_name = True
        self.display = pg.Surface(size)

    def load_font(self, font_path: str) -> None:
        if not isinstance(font_path, str) or not os.path.exists(font_path): return
        del self.font
        self.font_path = font_path
        self.font: pg.Font = pg.Font(font_path, self.text_size)

    def set_element(self, key: str, element: Element) -> None:
        self.elements[key] = element
    
    def get_element(self, key: str) -> Element:
        return self.elements.get(key, None)
    
    def rem_element(self, key: str) -> None:
        if self.get_element(key) is not None:
            del self.elements[key]

    def set_text_field(self, field: str, text: str, color: list[int]=None) -> bool:
        try:
            self.text_fields[field] = {"text": text, "color": color}
            return True
        except KeyError as e:
            return False
    
    def rem_text_field(self, field: str) -> bool:
        try:
            del self.text_fields[field]
            return True
        except KeyError as e:
            return False

    def render(self) -> None:
        if self.show_name: self.window.window.blit(self.font.render(self.name, True, self.title_color), self.location)
       
        for index, field in enumerate(self.text_fields.keys()): # render text fields
            text = f"{field}: {self.text_fields[field]["text"]}"
            color = self.text_fields[field]["color"]
            text_surface = self.font.render(text, True, self.text_color if not color else color)
            text_location = [
                self.location[0],
                self.location[1] + (text_surface.get_size()[1] * (index + 1))
            ]
            self.window.window.blit(text_surface, text_location)

        for element in self.elements.values():    # render elements
            element.render(self.window.window)

    def update(self, event_manager) -> None:
        for element in self.elements.values():
            element.update(event_manager)
            mouse_within = point_inside(Mouse.location.screen, [
                element.location[0] - element.border_size[0], element.location[1] - element.border_size[1],
                element.size[0] + element.border_size[0], element.size[1] + element.border_size[1]
            ])
            if not element.get_state(ELEMENT_STATE.HOVERED) and mouse_within:
                Mouse.Hovering = Element
                element.set_state(ELEMENT_STATE.HOVERED)
                element.on_hover()
            if element.get_state(ELEMENT_STATE.HOVERED) and not mouse_within:
                Mouse.Hovering = None
                element.rem_state(ELEMENT_STATE.HOVERED)
                element.on_unhover()
            if element.get_state(ELEMENT_STATE.HOVERED) and event_manager.mouse_pressed(Mouse.LeftClick):
                event_manager.mouse[Mouse.LeftClick] = 0    # shouldnt need this but fixes the element double-click issue :|
                element.on_click()
# ------------------------------------------------------------ #



R3FRAME_QUOTES: list[str] = [
    "howdy gamers!",
    "its all about perspective...",
    "now with colors!",
    "simple interfaces, complex tasks...",
    "made with <3",
    "what will you create?",
    "release date? when?",
    "is it... stable yet?",
    "the beauty is in the design!"
]



# ------------------------------------------------------------ #
class Atom:
    def __init__(self, aid: int, flags: int) -> None:
        self._data = [aid, flags]

    def ID(self) -> int:    return self._data[0]
    def FLAGS(self) -> int: return self._data[1]

    def set_state(self, flag: int) -> None:
        self._data[1] |= flag

    def get_state(self, flag: int) -> bool:
        return ((self._data[1] & flag) == flag)

    def rem_state(self, flag: int) -> None:
        self._data[1] &= ~flag
# ------------------------------------------------------------ #



from r3frame.globs import os, math

sine_wave_value = lambda A, B, t, C, D: int(A * math.sin((B * t) + C) + D)
"""
@brief Computes the y-coordinate of an oscillating motion following a sinusoidal (oscillatory, periodic, circular) wave.
@param A: Amplitude - The maximum height of the wave.
@param B: Frequency - Controls how fast the wave oscillates.
@param t: Time - Drives the motion forward.
@param C: Phase shift - Moves the wave left or right in time.
@param D: Vertical shift - Raises or lowers the wave.

@details
The function models a point’s vertical position as it moves in a circular or oscillatory path.
Mathematically, it returns the y-value of a sine wave:
    y = A * sin(B * t + C) + D
- The sine function describes smooth periodic motion.
- A larger amplitude (A) makes the oscillation taller.
- A higher frequency (B) makes it oscillate faster.
- The phase shift (C) offsets the wave horizontally.
- The vertical shift (D) moves the entire wave up/down.
"""

damp_exp = lambda x, factor, dt: x * (factor ** dt) if abs(x) > 0.01 else 0
"""
@brief Applies exponential damping to smoothly reduce a value over time.
@param x: The value to be damped (e.g., velocity, intensity, etc.).
@param factor: Damping factor (0 < factor < 1). Smaller values cause faster decay.
@param dt: Delta time (time step for decay application).

@details
This function simulates smooth, natural decay by multiplying `x` by a fractional factor:
    x_new = x * (factor ** dt)
- Works well for reducing velocity, smoothing animations, or applying drag-like effects.
- Ensures small values snap to 0 when they are close to stopping.
"""

damp_lin = lambda x, rate, threshold, dt: 0 if abs(x) < threshold else x - rate * dt * (1 if x > 0 else -1)
"""
@brief Applies linear damping to reduce a value at a fixed rate over time.
@param x: The value to be damped (e.g., velocity).
@param rate: The amount to reduce per second.
@param threshold: The stopping threshold (if |x| < threshold, it snaps to 0).
@param dt: Delta time (time step for decay application).

@details
This function simulates a linear friction effect by reducing `x` by a constant amount:
    x_new = x - rate * dt * sign(x)
- Good for simulating dry friction, velocity decay, or gradual slowdowns.
- Ensures the value stops completely when it gets close to 0.
"""

# ------------------------------------------------------------ #
div_v2 = lambda v, s: [v[0] / s, v[1] / s]
dist_v2 = lambda a, b: mag_v2(sub_v2(a, b))
scale_v2 = lambda v, s: [v[0] * s, v[1] * s]
mag_v2 = lambda v: (v[0]**2 + v[1]**2) ** 0.5
add_v2 = lambda a, b: [a[0] + b[0], a[1] + b[1]]
sub_v2 = lambda a, b: [a[0] - b[0], a[1] - b[1]]
clamp = lambda v, l, u: l if v < l else u if v > u else v
norm_v2 = lambda v: [v[0] / mag_v2(v), v[1] / mag_v2(v)] if mag_v2(v) != 0 else [0, 0]
equal_arrays = lambda a, b: all([*map(lambda a, b: a == b, a, b)])
unequal_arrays = lambda a, b: all([*map(lambda a, b: a != b, a, b)])
# ------------------------------------------------------------ #

def abs_path(path: str) -> str:
    fp = __file__.split(os.sep)
    fp.remove(fp[len(fp)-1])
    [fp.append(p) for p in path.replace("/", os.sep).replace("\\", os.sep).split(os.sep)]
    fp = f"{os.sep}".join(fp)
    return fp

def rel_path(path: str) -> str:
    fp = path.replace("/", os.sep).replace("\\", os.sep)
    return fp

def point_inside(point: list[int|float], bounds: list[int|float]) -> bool:
    return \
        point[0] > bounds[0] and point[0] < bounds[0] + bounds[2] \
    and point[1] > bounds[1] and point[1] < bounds[1] + bounds[3]

def angle_to(from_point: list[int|float], to_point: list[int|float]) -> float:
    """Calculates the angle (in degrees) between two points."""
    return math.degrees(math.atan2(*[
        to_point[0] - from_point[0],
        to_point[1] - from_point[1]
    ])) 

def bsort(data: list[int]) -> list[int]:
    for i in range(len(data)-1, 0, -1):
        for j in range(i):
            if data[j] > data[j+1]:
                temp = data[j]
                data[j] = data[j+1]
                data[j+1] = temp



from r3frame.globs import pg
from r3frame.atom import Atom
from r3frame.util import mag_v2, div_v2, sub_v2, clamp

# ------------------------------------------------------------ #
class Object(Atom):
    def __init__(
            self,
            pos: list[int],
            size: list[int],
            color: list[int],
            bounds: list[int]=None,
    ) -> None:
        super().__init__(0, 0)
        self.data = [
            size[:], color[:], pg.Surface(size),
            100, pos[:], [0, 0, 0, 0, None, bounds[:] if bounds is not None else None]] 
        """ `Object` data array: [size, color, surface, speed, pos, movement[left, right, up, down, move_to, bounds]] """
        self.fill()

    @property
    def size(self) -> list[int]:        return self.data[0]
    @property
    def color(self) -> list[int]:       return self.data[1]
    @property
    def image(self) -> pg.Surface:      return self.data[2]
    @property
    def speed(self) -> float:           return self.data[3]
    @property
    def pos(self) -> list[float]:       return self.data[4]
    @property
    def movement(self) -> list[bool]:   return self.data[5]
    @property
    def bounds(self) -> float:          return self.data[5][5]
    @property
    def colorkey(self) -> None: return self.data[2].get_colorkey()
    @property
    def center(self) -> list[float]:
        return [self.data[4][0] + (self.data[0][0] / 2),
                self.data[4][1] + (self.data[0][1] / 2)]
    
    def fill(self) -> None:
        self.data[2].fill(self.data[1])
    
    def mod_color(self, color: list[int]) -> None:
        self.data[1] = color[:]
    
    def mod_colorkey(self, color: list[int]) -> None:
        self.data[2].set_colorkey(color)
    
    def mod_bounds(self, bounds: list[int]) -> None:
        self.data[5][5][0] = bounds[0]
        self.data[5][5][1] = bounds[1]

    def rect(self) -> pg.Rect:
        return pg.Rect(self.data[4], self.data[0])
    
    def move_to(self, pos: list[float]) -> None:
        self.data[5][4] = pos[:] if pos else None

    def move(self, l: bool=None, r: bool=None, u: bool=None, d: bool=None) -> None:
        self.data[5][0] = True if l else False
        self.data[5][1] = True if r else False
        self.data[5][2] = True if u else False
        self.data[5][3] = True if d else False
    
    def _aabbx(self, transform: list[float], collidables: list["Object"]) -> None:
        rect = self.rect()
        for obj in collidables:
            if rect.colliderect(obj.rect()):
                if transform[0] < 0:
                    rect.left = obj.rect().right
                    self.data[5][0] = 0
                if transform[0] > 0:
                    rect.right = obj.rect().left
                    self.data[5][1] = 0
                self.data[4][0] = rect.x
    
    def _aabby(self, transform: list[float], collidables: list["Object"]) -> None:
        rect = self.rect()
        for obj in collidables:
            if rect.colliderect(obj.rect()):
                if transform[1] < 0:
                    rect.top = obj.rect().bottom
                    self.data[5][2] = 0
                if transform[1] > 0:
                    rect.bottom = obj.rect().top
                    self.data[5][3] = 0
                self.data[4][1] = rect.y

    def update(self, dt: float, collidables: list["Object"]=[]) -> None:
        if self.data[5][4]:
            diff = sub_v2(self.data[5][4], self.center)
            dist = mag_v2(diff)
            if int(dist) <= 8:
                self.data[5][4] = None
            else:
                dirt = div_v2(diff, dist)
                if self.data[5][5]:
                    self.data[4][0] = clamp(self.data[4][0] + dirt[0] * self.data[3] * dt, 0, self.data[5][5][0])
                    self._aabbx(dirt, collidables)
                    self.data[4][1] = clamp(self.data[4][1] + dirt[1] * self.data[3] * dt, 0, self.data[5][5][1])
                    self._aabby(dirt, collidables)
                else:
                    self.data[4][0] += dirt[0] * self.data[3] * dt
                    self._aabbx(dirt, collidables)
                    self.data[4][1] += dirt[1] * self.data[3] * dt
                    self._aabby(dirt, collidables)
        else:        
            t = [(self.data[5][1] - self.data[5][0]) * self.data[3],
                (self.data[5][3] - self.data[5][2]) * self.data[3]]
            if self.data[5][5]:
                self.data[4][0] = clamp(self.data[4][0] + t[0] * dt, 0, self.data[5][5][0])
                self._aabbx(t, collidables)
                self.data[4][1] = clamp(self.data[4][1] + t[1] * dt, 0, self.data[5][5][1])
                self._aabby(t, collidables)
            else:
                self.data[4][0] += t[0] * dt
                self._aabbx(t, collidables)
                self.data[4][1] += t[1] * dt
                self._aabby(t, collidables)
# ------------------------------------------------------------ #


from r3frame.globs import time

# ------------------------------------------------------------ #
class Clock:
    FPS:int=0
    maxFPS:int=60
    last:float=0.0
    delta:float=0.0
    current:float=0.0

    def update(self) -> None:
        self.current = time.time()

        if self.last == 0.0:
            self.delta = 0.0
        else: self.delta = self.current - self.last

        self.last = self.current

        if self.delta > 0: self.FPS = 1 / self.delta

    def rest(self) -> None:
        time.sleep(max(1 / self.maxFPS - self.delta, 0))
# ------------------------------------------------------------ #



from r3frame.globs import os, re, pg, json, random
from r3frame.app.resource.image import Image
from r3frame.util import unequal_arrays
from r3frame.game.obj import Object

# ------------------------------------------------------------ #
class Tilemap(Object):
    def __init__(self, app, size: list[int], tilesize: int=32, grid_color: list[int]=[10, 10, 10]) -> None:
        self._ = pg.Surface([tilesize, tilesize])
        self._.fill([255, 255, 255])
        self.app = app
        self.sizet = size                                                                       # in tiles
        self.width = size[0]                                                                    # in tiles
        self.height = size[1]                                                                   # in tiles
        self.tilesize = tilesize                                                                # in pixels
        self.sizep: list[int] = [self.sizet[0] * self.tilesize, self.sizet[1] * self.tilesize]  # in pixels

        self.tilesets: list[Image] = []
        self.layers: dict[str, list] = {
            "background": [[None for _ in range(self.sizet[0] * self.sizet[1])] for _ in range(2)],
            "foreground": [[None for _ in range(self.sizet[0] * self.sizet[1])] for _ in range(2)],
            "midground": [[None for _ in range(self.sizet[0] * self.sizet[1])] for _ in range(2)]
        }

        self.grid_color: list[int] = grid_color

        super().__init__([0, 0], self.sizep, [255, 255, 255])

    def export_image(self, name: str, path: str, grid: bool=False) -> None:
        if not os.path.exists(path): return False
        for i in range(len(self.tilesets)):
            for layer in self.layers:
                self.read_data(layer, i, self.sizet, self.tilesize, self.layers[layer][0])
        if grid: self.render_grid()
        pg.image.save(self.image, os.path.join(path, f"{name}.png"))

    def read_data(self, layer: str, tileset: int, sizet: list[int], tilesize:int, data: list[int|str]) -> None:
        if tileset < 0 or tileset >= len(self.tilesets): return
        if not self.layers.get(layer, False): return
        for x in range(sizet[0]):
            for y in range(sizet[1]):
                data_tile = data[y * sizet[0] + x]
                if data_tile != -1 and data_tile != None:
                    self.set_tile(layer, tileset, data_tile, [x * tilesize, y * tilesize])

    def export_data(self, path: str) -> bool:
        if not os.path.exists(path): return False

        tilesets = []
        for ts in self.tilesets:
            for tileset in ts:
                if len(tilesets) and tileset.path == tilesets[len(tilesets)-1]:
                    continue
                tilesets.append(tileset.path)
        
        save_data = {
            "config": {
                "tilesize": self.tilesize,
                "sizet": self.sizet,
                "tilesets": tilesets
            },
            "layers": {}
        }

        for layer in self.layers:
            save_data["layers"][layer] = {}
            for tile in self.layers[layer][1]:
                if tile is None: continue
                location = [
                    tile.location[0] // self.tilesize,
                    tile.location[1] // self.tilesize
                ]
                index = location[1] * self.sizet[0] + location[0]
                location = f"{location[0]},{location[1]}"
                save_data["layers"][layer][location] = self.layers[layer][0][index]
                
        with open(path, "w") as save:
            try:
                json.dump(save_data, save, indent=4, separators=(",", ": "))
            except json.JSONDecodeError as e: return False
        return True
    
    def import_data(self, path:str) -> bool:
        if not os.path.exists(path): return False
        with open(path, "r") as save:
            data = json.load(save)
            
            config: dict = data["config"]
            layers: dict = data["layers"]

            sizet = config["sizet"]
            tilesize = config["tilesize"]
            
            tilesets = config["tilesets"]
            for i, tileset in enumerate(tilesets):
                if not os.path.exists:
                    print(f"failed to load tileset: {tileset}")
                    return False
                self.tilesets.append(self.app.resource.load_image_sheet(
                    f"tileset{i}", tileset, [tilesize, tilesize]
                ))

            for layer in layers:
                for location, data in layers[layer].items():
                    location = [*map(int, location.split(","))]
                    if data[0] != -1 and data[0] != None:
                        self.set_tile(layer, data[0], data[1], [location[0] * tilesize, location[1] * tilesize])
        return True
    
    def import_tileset(self, path: str) -> bool:
        if not os.path.exists(path): return False
        self.tilesets.append(self.app.resource.load_image_sheet(f"tileset{len(self.tilesets)-1}", path, [self.tilesize, self.tilesize]))
        return True

    def set_data(self, layer: str, tileset: int, tile_id: int, location: list[int]) -> None:
        if tileset < 0 or tileset >= len(self.tilesets): return
        if not self.layers.get(layer, False): return
        location = [
            int(location[0] - self.location[0]),
            int(location[1] - self.location[1])
        ]
        mapx = location[0] // self.tilesize
        mapy = location[1] // self.tilesize
        if mapx < 0 or mapy < 0 or mapx >= self.sizet[0] or mapy >= self.sizet[1]: return
        self.layers[layer][0][mapy * self.sizet[0] + mapx] = [tileset, tile_id]
    
    def rem_data(self, layer: str, location: list[int]) -> None:
        if not self.layers.get(layer, False): return
        location = [
            int(location[0] - self.location[0]),
            int(location[1] - self.location[1])
        ]
        mapx = location[0] // self.tilesize
        mapy = location[1] // self.tilesize
        if mapx < 0 or mapy < 0 or mapx >= self.sizet[0] or mapy >= self.sizet[1]: return
        self.layers[layer][0][mapy * self.sizet[0] + mapx] = None

    def get_data(self, layer: str, location: list[int]) -> int|str|None:
        if not self.layers.get(layer, False): return
        location = [
            int(location[0] - self.location[0]),
            int(location[1] - self.location[1])
        ]
        mapx = location[0] // self.tilesize
        mapy = location[1] // self.tilesize
        if mapx < 0 or mapy < 0 or mapx >= self.sizet[0] or mapy >= self.sizet[1]: return None
        return self.layers[layer][0][mapy * self.sizet[0] + mapx]
    
    def get_tile(self, layer: str, location: list[int]) -> int:
        if not self.layers.get(layer, False): return
        location = [
            int(location[0] - self.location[0]),
            int(location[1] - self.location[1])
        ]
        mapx = location[0] // self.tilesize
        mapy = location[1] // self.tilesize
        if mapx < 0 or mapy < 0 or mapx >= self.sizet[0] or mapy >= self.sizet[1]: return None
        return self.layers[layer][1][mapy * self.sizet[0] + mapx]

    def set_tile(self, layer: str, tileset: int, tile_id: int, location: list[int]) -> None:
        if tileset < 0 or tileset >= len(self.tilesets): return
        if not self.layers.get(layer, False): return
        if self.get_tile(layer, location) is not None: return
        
        location = [
            int(location[0] - self.location[0]),
            int(location[1] - self.location[1])
        ]

        mapx = (location[0] * self.tilesize) // self.tilesize
        mapy = (location[1] * self.tilesize) // self.tilesize
        mapx -= mapx % self.tilesize
        mapy -= mapy % self.tilesize
        if (
            mapx < 0 or mapx > self.size[0] or mapx + self.tilesize > self.size[0] or
            mapy < 0 or mapy > self.size[1] or mapy + self.tilesize > self.size[1]
            ): return
        
        tile = Object(
            size=[self.tilesize, self.tilesize], color=[
                random.randint(0, 255),
                random.randint(0, 255),
                random.randint(0, 255)
            ], location=[
                (location[0] // self.tilesize) * self.tilesize,
                (location[1] // self.tilesize) * self.tilesize
            ]
        )

        tile.id = tile_id

        if tile_id < 0 or tile_id >= len(self.tilesets[tileset]): pass
        else:
            tile.image = self.tilesets[tileset][tile_id]
            self.image.blit(tile.image.data, [mapx, mapy])
        
        index = location[1] // self.tilesize * self.sizet[0] + location[0] // self.tilesize
        self.layers[layer][0][index] = [tileset, tile_id]
        self.layers[layer][1][index] = tile

    def rem_tile(self, layer: str, location: list[int]) -> None:
        if not self.layers.get(layer, False): return
        if self.get_tile(layer, location) is None: return
        location = [
            int(location[0] - self.location[0]),
            int(location[1] - self.location[1])
        ]

        mapx = (location[0] * self.tilesize) // self.tilesize
        mapy = (location[1] * self.tilesize) // self.tilesize
        mapx -= mapx % self.tilesize
        mapy -= mapy % self.tilesize
        if (
            mapx < 0 or mapx > self.size[0] or mapx + self.tilesize > self.size[0] or
            mapy < 0 or mapy > self.size[1] or mapy + self.tilesize > self.size[1]
            ): return
        
        self.image.blit(self._, [mapx, mapy])
        index = location[1] // self.tilesize * self.sizet[0] + location[0] // self.tilesize
        self.layers[layer][0][index] = None
        self.layers[layer][1][index] = None

    def _generate_region(self, size:list[int], location:list[int]) -> list[list[int]]:
        center = [
            int(location[0] // self.tilesize),
            int(location[1] // self.tilesize)
        ]; region = []
        for x in range(center[0] - size[0], (center[0] + size[0]) + 1):
            for y in range(center[1] - size[1], (center[1] + size[1]) + 1):
                region.append([x, y])
        return region

    def get_region(self, layer:str, size:list[int], location:list[int]) -> list[Object]|None:
        if not self.layers.get(layer, False): return
        region = self._generate_region(size, location)
        if not region: return None
        tiles = []
        for map_location in region:
            index = map_location[1] * self.sizet[0] + map_location[0]
            if index < 0 or index >= (self.sizet[0] * self.sizet[1]): continue
            tile = self.layers[layer][1][index]
            if tile: tiles.append(tile)
        return tiles

    def render_grid(self) -> None:
        start = [0, 0]
        end = [
            (start[0] + (self.sizet[0] * self.tilesize)) // self.tilesize,
            (start[1] + (self.sizet[1] * self.tilesize)) // self.tilesize
        ]
        
        for gx in range(int(start[0]), int(end[0])):
            x = gx * self.tilesize
            pg.draw.line(self.image, self.grid_color, [x, start[1] * self.tilesize], [x, end[1] * self.tilesize], 1)
        
        for gy in range(int(start[1]), int(end[1])):
            y = gy * self.tilesize
            pg.draw.line(self.image, self.grid_color, [start[0] * self.tilesize, y], [end[0] * self.tilesize, y], 1)
# ------------------------------------------------------------ #



from r3frame.globs import pg

# ------------------------------------------------------------ #
class StaticPartition:
    def __init__(self, size: list[int], cellsize: int) -> None:
        self.objs = []
        self.size = size                                                # in cells
        self.width = size[0]                                            # in cells
        self.height = size[1]                                           # in cells
        self.cellsize = cellsize                                       # in pixels
        self.cells = [[] for _ in range(self.width * self.height)]

    def get_cell(self, location: list[int]) -> list:
        """Returns all objects at the given world (pixel) position."""
        gx = int(location[0]) // self.cellsize
        gy = int(location[1]) // self.cellsize
        if gx < 0 or gy < 0 or gx > self.size[0] or gy > self.size[1]: return
        index = gy * self.width + gx
        if index >= len(self.cells): return
        return self.cells[index]
    
    def set_cell(self, obj) -> None:
        """Places an object in the grid at the given world (pixel) position."""
        if isinstance(obj, list):
            for o in obj:
                if o is None: continue
                cell = self.get_cell(o.pos)
                if cell is None or o in cell: return
                self.objs.append(o)
                cell.append(o)
        else:
            cell = self.get_cell(obj.pos)
            if cell is None or obj in cell: return
            self.objs.append(obj)
            cell.append(obj)

    def rem_cell(self, obj) -> None:
        """Removes and an object from the grid at the given world (pixel) position."""
        cell = self.get_cell(obj.pos)
        if cell is None: return
        self.objs.remove(obj)
        cell.remove(obj)

    def _generate_region(self, size:list[int], location:list[int]) -> list[list[int]]:
        center = [
            int(location[0]) // self.cellsize,
            int(location[1]) // self.cellsize
        ]; region = []
        for x in range(center[0] - size[0], (center[0] + size[0]) + 1):
            for y in range(center[1] - size[1], (center[1] + size[1]) + 1):
                region.append([x, y])
        return region
    
    def get_area(self, topleft:list[int], bottomright:list[int]) -> list[list[int]]:
        start = [
            int(topleft[0]) // self.cellsize,
            int(topleft[1]) // self.cellsize
        ]; end = [
            int(bottomright[0]) // self.cellsize,
            int(bottomright[1]) // self.cellsize
        ]; cells = []
        for x in range(start[0], end[0]):
            for y in range(start[1], end[1]):
                cell = self.cells[y * self.size[0] + x]
                if cell:
                    cells.append(cell)
        return [o for cell in cells for o in cell if cell]

    def get_region(self, size:list[int], location:list[int]) -> list|None:
        region = self._generate_region(size, location)
        if not region: return None
        cells = []
        for map_location in region:
            index = map_location[1] * self.size[0] + map_location[0]
            if index >= len(self.cells): continue
            cell = self.cells[index]
            if cell: cells.append(cell)
        return cells

    def debug_render(self, renderer, center: list[int|float], radius: int=8) -> None:
        """Renders a portion of the grid around a center position within a given radius."""
        cell_w, cell_h = self.cellsize, self.cellsize
        radius_px = radius * cell_w  # convert radius from cells to pixels

        # convert world position to grid cell indices
        start_x = max((int(center[0]) - radius_px) // cell_w, 0)
        start_y = max((int(center[1]) - radius_px) // cell_h, 0)
        end_x = min((int(center[0]) + radius_px) // cell_w, self.size[0] - 1)
        end_y = min((int(center[1]) + radius_px) // cell_h, self.size[1] - 1)

        # draw grid lines within the region
        for gx in range(start_x, end_x):
            x = gx * cell_w
            renderer.window.draw_line([x, start_y * cell_h], [x, end_y * cell_h], [125, 125, 125], 1)
        
        for gy in range(start_y, end_y):
            y = gy * cell_h
            renderer.window.draw_line([start_x * cell_w, y], [end_x * cell_w, y], [125, 125, 125], 1)

        # draw objects within the visible region
        for obj in self.objs:
            x, y = obj.center
            gx = x // self.cellsize
            gy = y // self.cellsize
            if start_x <= gx <= end_x and start_y <= gy <= end_y:
                obj_x, obj_y = gx * cell_w, gy * cell_h
                renderer.window.draw_circle([obj_x + self.cellsize / 2, obj_y + self.cellsize / 2], self.cellsize, [0, 255, 0], 1)
# ------------------------------------------------------------ #


from r3frame.globs import pg
from r3frame.app.camera import Camera
from r3frame.app.resource.image import Image

# ------------------------------------------------------------ #
class Renderer:
    """
    Handles rendering of objects to the display, with support for different rendering strategies 
    optimized for small and large game worlds.
    """
    
    class FLAGS:
        SHOW_CAMERA: int = 1 << 0  # flag to display the camera's viewport boundaries.

    def __init__(self, camera: Camera) -> None:
        """
        Initializes the renderer with a target window and camera.

        :param window: The game window where rendering occurs.
        :param camera: The camera that defines the viewport.
        """
        self.camera = camera
        self.window = camera.window
        self.target = self.window.display
        self.flags = 0
        self.draw_calls = 0
        self._draw_calls = []  # draw_call layout : [image, location]

    def set_flag(self, flag: int) -> None:
        """Enables a rendering flag."""
        self.flags |= flag

    def rem_flag(self, flag: int) -> None:
        """Disables a rendering flag."""
        if (self.flags & flag) == flag:
            self.flags &= ~flag

    def pre_render(self) -> None: pass
    
    def post_render(self) -> None: pass

    def draw_call(self, image: Image, location: list[int]) -> None:
        """
        Queues a draw call for rendering.

        :param surface: The image/surface to render.
        :param location: The world-space position of the surface.
        
        Performs frustum culling to avoid rendering objects outside of the viewport.
        """
        if self.draw_calls + 1 > 4096:  # prevent excessive draw calls.
            return

        # frustum culling
        if ((location[0] + image.size[0]) - self.window.clip_range[0] < self.camera.location[0] or 
            location[0] + self.window.clip_range[0] > self.camera.location[0] + self.camera.viewport_size[0]) or \
           ((location[1] + image.size[1]) - self.window.clip_range[1] < self.camera.location[1] or 
            location[1] + self.window.clip_range[1] > self.camera.location[1] + self.camera.viewport_size[1]):
            return

        self._draw_calls.append([image, location])
        self.draw_calls += 1

    def render(self) -> None:
        """
        Renders objects to a viewport-sized surface before scaling it up to the display.
        
        Solves the **display transformation bottleneck** by limiting rendering to objects within 
        the camera's viewport, making it efficient for **large game worlds**. Since transformations 
        are applied at the render-target/display level, object positions remain true to world coordinates.
        """
        del self.target
        self.target = pg.Surface(self.camera.viewport_size)  # create a surface matching the viewport size.
        self.target.fill(self.window.color)
        self.window.clear()

        self.pre_render()
        for i in range(self.draw_calls):
            image, location = self._draw_calls.pop(0)
            self.window.blit(image, location)
        self.draw_calls = 0
        self.post_render()

        if (self.flags & self.FLAGS.SHOW_CAMERA):
            self.window.blit_rect(self.camera.get_viewport(), color=[255, 255, 255], width=1)
            self.window.blit_rect(self.camera.get_center([10, 10]), color=[0, 255, 0], width=1)
        
        # apply camera transformations at the render-target level (no per-object transformations)
        self.target.blit(self.window.display, [-self.camera.location[0], -self.camera.location[1]])
        self.window.window.blit(
            pg.transform.scale(self.target, self.window.size),
            [0, 0]
        )
# ------------------------------------------------------------ #



from r3frame.globs import pg
from r3frame.app.resource.base import Resource

# ------------------------------------------------------------ #
class Animation(Resource):
    def __init__(
            self,
            rid: int,
            frames: list[pg.Surface],
            loop: bool=1, frame_duration: float=5.0,
            frame_offset: list[int]=[0, 0]
        ) -> None:
        super().__init__(rid, frames)
        self.done = 0
        self.frame = 0
        self.loop = loop
        self.flip_x = False
        self.flip_y = False
        self.frame_offset = frame_offset
        self.frame_duration = frame_duration

    def reset(self) -> None: self.frame, self.done = 0, 0

    def copy(self):
        return Animation(self.data, self.loop, self.frame_duration, self.frame_offset)

    def get_frame(self):
        return pg.transform.flip(self.data[int(self.frame / self.frame_duration)].data, self.flip_x, self.flip_y)

    def update(self) -> None:
        if self.loop:
            self.frame = (self.frame + 1) % (self.frame_duration * len(self.data))
        else:
            self.frame = min(self.frame + 1, self.frame_duration * len(self.data) - 1)
            if self.frame >= self.frame_duration * len(self.data) - 1:
                self.done = 1
# ------------------------------------------------------------ #



from r3frame.globs import pg
from r3frame.atom import Atom

# ------------------------------------------------------------ #
class VFX(Atom):
    def __init__(self, id: int, app) -> None:
        super().__init__()
        self.id: int = id
        self.app = app

    def update_render(self, surface: pg.Surface) -> None: raise NotImplementedError
# ------------------------------------------------------------ #



from r3frame.globs import pg
from r3frame.util import damp_lin
from r3frame.app.window import Window

# ------------------------------------------------------------ #
class Camera:
    class MODES:
        CENTER_ON: int = 1

    def __init__(self, window: Window):
        self.window = window
        self.mode = 0
        self.drag = 18
        self.speed = 100
        self.location = [0, 0]
        self.velocity = [0.0, 0.0]
        self.last_location = self.location

        self.bounds = window.display_size
        self.viewport_size = window.display_size
        self.viewport_scale = [
            self.window.size[0] / self.viewport_size[0],
            self.window.size[1] / self.viewport_size[1]
        ]
        self.center = [self.location[0] + self.viewport_size[0] / 2, self.location[1] + self.viewport_size[1] / 2]
        self.mod_viewport(-self.viewport_size[0] - self.viewport_size[1])

    def configure(self, display_size: list[int]) -> None:
        self.bounds = display_size
        self.viewport_size = display_size
        self.viewport_scale = [
            self.window.size[0] / self.viewport_size[0],
            self.window.size[1] / self.viewport_size[1]
        ]
        self.center = [self.location[0] + self.viewport_size[0] / 2, self.location[1] + self.viewport_size[1] / 2]
        self.mod_viewport(-self.viewport_size[0] - self.viewport_size[1])

    def get_center(self, size: list[int]) -> pg.Rect:
        return pg.Rect([self.center[0] - size[0] / 2, self.center[1] - size[1] / 2], size)

    def get_viewport(self) -> pg.Rect:
        return pg.Rect(self.location, self.viewport_size)

    def set_velocity(self, vx: float=0.0, vy: float=0.0) -> None:
        if vx: self.velocity[0] = vx
        if vy: self.velocity[1] = vy

    def mod_viewport(self, delta: float) -> list[int]:
        delta *= (min(self.viewport_size) * 0.05)  # scale the delta by 5% of the viewport size
        aspect_ratio = self.viewport_size[0] / self.viewport_size[1]

        new_width = min(self.bounds[0], max(260, self.viewport_size[0] + delta))
        new_height = min(self.bounds[1], max(260, self.viewport_size[1] + delta))

        if new_width / new_height != aspect_ratio:
            if new_width == self.bounds[0]:
                new_height = new_width / aspect_ratio
            if new_height == self.bounds[1]:
                new_width = new_height * aspect_ratio

        self.viewport_size = [new_width, new_height]
        self.center = [self.location[i] + self.viewport_size[i] / 2 for i in (0, 1)]

        return self.viewport_size

    def center_on(self, size: list[int], location: list[int|float]) -> None:
        if self.mode != self.MODES.CENTER_ON: self.mode = self.MODES.CENTER_ON
        target_center = [
            (location[0] + self.viewport_size[0] / 2) + size[0] / 2,
            (location[1] + self.viewport_size[1] / 2) + size[1] / 2
        ]

        dist = [
            (self.center[0] - target_center[0]) + self.viewport_size[0] / 2,
            (self.center[1] - target_center[1]) + self.viewport_size[1] / 2
        ]
    
        self.velocity = [
            (-dist[0] * self.speed) * (1 / self.drag),
            (-dist[1] * self.speed) * (1 / self.drag)
        ]

    def update(self, delta_time: float) -> None:
        self.viewport_scale = [
            self.window.size[0] / self.viewport_size[0],
            self.window.size[1] / self.viewport_size[1]
        ]
        self.last_location = self.location
        self.velocity = [damp_lin(v, self.speed, 3, delta_time) for v in self.velocity]
        self.location[0] = max(0, min(self.bounds[0] - self.viewport_size[0], self.location[0] + self.velocity[0] * delta_time))
        self.location[1] = max(0, min(self.bounds[1] - self.viewport_size[1], self.location[1] + self.velocity[1] * delta_time))
        self.center = [self.location[0] + self.viewport_size[0] / 2, self.location[1] + self.viewport_size[1] / 2]
# ------------------------------------------------------------ #


from r3frame.globs import pg
from r3frame.util import abs_path

from r3frame.atom import Atom
from r3frame.app.proc import Process
from r3frame.app.scene import Scene
from r3frame.app.clock import Clock
from r3frame.app.window import Window
from r3frame.app.camera import Camera
from r3frame.app.renderer import Renderer
from r3frame.app.event import EventManager
from r3frame.app.input import Keyboard, Mouse
from r3frame.app.resource.image import Image
from r3frame.app.resource.manager import ResourceManager

# ------------------------------------------------------------ #
class Application:
    def __init__(self, name: str="My App", window_size: list[int]=[800, 600]) -> None:
        self.state: int = 0
        self.name:str = name
        self.clock: Clock = Clock()
        self.events: EventManager = EventManager()
        self.resource: ResourceManager = ResourceManager()

        self.scene: str = None
        self.scene: Scene = None
        self.scenes: dict[str, Scene] = {}

        self.window: Window = Window(window_size, window_size)
        self.window.title = name
        self.window.icon = pg.image.load(abs_path("assets/images/r3-logo.ico"))
        self.window.configure(window_size)

        self.camera: Camera = Camera(self.window)
        self.renderer: Renderer = Renderer(self.camera)

        self.processes: list[Process] = [None for _ in range(100)]
        self.configure()
    
    def _bsort(self) -> None:
        for _ in range(len(self.processes) - 1, 0, -1):
            for __ in range(_):
                if not self.processes[__] or not self.processes[__+1]: continue
                if self.processes[__].id > self.processes[__+1].id:
                    t = self.processes[__]
                    self.processes[__] = self.processes[__+1]
                    self.processes[__+1] = t
        
    def add_proc(self, proc: Process) -> None:
        if not isinstance(proc, Process): return
        self.processes[proc.id] = proc

    def rem_proc(self, pid: int) -> None:
        try:
            self.processes.pop(pid)
        except IndexError: pass

    def set_scene(self, scene: Scene) -> None:
        self.scenes[scene.name] = scene
        self.scene = self.scenes[scene.name]
        self.window.configure(self.scene.display_size)
        self.camera.configure(self.scene.display_size)
    
    def rem_scene(self, key: str) -> Scene|None:
        if self.get_scene(key) is not None:
            self.scenes.pop(key, None)
            self.scene = None
    
    def get_scene(self, key: str) -> Scene|None:
        return self.scenes.get(key, None)

    def configure(self) -> None: raise NotImplementedError
    def handle_events(self) -> None: raise NotImplementedError
    def handle_update(self) -> None: raise NotImplementedError
    def handle_render(self) -> None: raise NotImplementedError

    def exit(self) -> None: raise NotImplementedError
    def run(self) -> None:
        while not self.events.quit:
            self.clock.update()
            self.events.update()

            if isinstance(self.scene, Scene):
                self.scene.handle_events()
                self.handle_events()

                self.scene.handle_update()
                self.scene.interface.update(self.events)
                self.handle_update()
                self.camera.update(self.clock.delta)

                self.scene.handle_render()
                self.handle_render()
                self.renderer.render()
                self.scene.interface.render()
                
                if isinstance(Mouse.cursor.image, Image):
                    self.window.window.blit(Mouse.cursor.image.data, Mouse.location.screen)

                Mouse.location.world = [
                    Mouse.location.view[0] // self.scene.tilemap.tilesize,
                    Mouse.location.view[1] // self.scene.tilemap.tilesize
                ]
            else:
                self.handle_events()

                self.handle_update()
                self.camera.update(self.clock.delta)

                self.handle_render()
                self.renderer.render()
                if isinstance(Mouse.cursor.image, Image):
                    self.window.window.blit(Mouse.cursor.image.data, Mouse.location.screen)

            Mouse.location.rel = [*pg.mouse.get_rel()]
            Mouse.location.screen = [*pg.mouse.get_pos()]
            Mouse.location.view = [
                int(Mouse.location.screen[0] // self.camera.viewport_scale[0] + self.camera.location[0]),
                int(Mouse.location.screen[1] // self.camera.viewport_scale[1] + self.camera.location[1]),
            ]
            self.window.update()
            self.clock.rest()
        else:
            self.exit()
# ------------------------------------------------------------ #



from r3frame.globs import os, pg
from r3frame.atom import Atom
from r3frame.util import equal_arrays
from r3frame.app.window import Window
from r3frame.app.resource.base import Resource

# ------------------------------------------------------------ #
class Image(Resource):
    def __init__(self, rid: int, path: str, data: pg.Surface, color: list[int]=[255, 255, 255]) -> None:
        super().__init__(rid, data)
        self.path: str = path
        self.rotation: float = 0.0
        self.color: list[int] = color
        self.scale: list[float] = [1.0, 1.0]
        self.size: list[int] = [self.data.get_width(), self.data.get_height()]

    def fill(self, color: list[int]=None) -> None:
        if not isinstance(color, list): return
        if equal_arrays(self.color, color): return
        self.data.fill(color if color else self.color)
        self.color = color if color else self.color

    def set_scale(self, scale: list[float]) -> None:
        if not isinstance(scale, list): return
        if equal_arrays(self.scale, scale): return
        self.data = pg.transform.scale(self.data, [
            self.size[0] * scale[0],
            self.size[1] * scale[1]
        ])
        self.scale = scale
# ------------------------------------------------------------ #


# ------------------------------------------------------------ #
class TextureConfig:
    def __init__(self, path: str, sizep: list[int]) -> None:
        self.path: str = path
        self.sizep: list[int] = sizep

class Atlas(Atom):
    def __init__(self, sizep: list[int]) -> None:
        super().__init__(0, 0)
        self.sizep: list[int] = sizep[:]
        self.location: list[int] = [0, 0]
        self.surface: pg.Surface = pg.Surface(sizep)
        self.surface.set_colorkey([1, 1, 1])
        self.surface.fill([1, 1, 1])
        self.data: list[tuple[str, list[int], list[int]]] = []

    def load_texture(self, config: TextureConfig) -> None:
        if not isinstance(config, TextureConfig): return
        if not os.path.exists(config.path): return
        tid = len(self.data)

        image = pg.image.load(config.path).convert_alpha()
        w, h = config.sizep

        if self.location[0] + w > self.sizep[0]:
            self.location[0] = 0
            self.location[1] += h

        if self.location[1] + h > self.sizep[1]: return

        self.surface.blit(image, self.location)
        del image

        self.data.append(([w, h], self.location[:]))
        self.location[0] += w  # move right for next texture

        return tid

    def unload_texture(self, tid: int) -> None:
        if tid < 0 or tid >= len(self.data): return
        
        size, location = self.data[tid]
        rect = pg.Rect(location, size)
        surface = pg.Surface(size)
        surface.fill([1, 1, 1])
        
        self.surface.blit(surface, location, rect)

    def blit_texture(self, tid: int, window: Window, blit_location: list[int], offset: list[int]=[0, 0]) -> None:
        if tid < 0 or tid >= len(self.data): return
        size, location = self.data[tid]
        rect = pg.Rect(location, size)
        window.blit(self.surface, blit_location, size, offset, rect)
# ------------------------------------------------------------ #



from r3frame.app.resource.base import Resource
from r3frame.app.resource.image import Image
from r3frame.app.resource.animation import Animation
from r3frame.app.resource.manager import ResourceManager


from r3frame.app.clock import Clock
from r3frame.app.proc import Process
from r3frame.app.window import Window
from r3frame.app.camera import Camera
from r3frame.app.renderer import Renderer
from r3frame.app.event import EventManager
from r3frame.app.input import Keyboard, Mouse

from r3frame.app.scene import Scene
from r3frame.app.base import Application

import r3frame.app.ui as ui
import r3frame.app.resource as resource


import setuptools

version = {
    "year" :2025,
    "minor" :0,
    "patch" :25
}

setuptools.setup(
    name='r3frame',
    version=f"{version["year"]}.{version["minor"]}.{version["patch"]}",
    description='A simple and powerful pygame framework',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    author='Izaiyah Stokes',
    author_email='d34d0s.dev@gmail.com',
    url='https://github.com/r3shape/r3frame',
    packages=setuptools.find_packages(),
    install_requires=[
        'pygame-ce'
    ],
    classifiers=[
        'Programming Language :: Python :: 3.12',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ], include_package_data=True
)


import r3frame.app as app
import r3frame.game as game
import r3frame.util as util
import r3frame.version as version
import r3frame.quotes as quotes

from r3frame.globs import os, random
if "R3FRAME_NO_PROMT" not in os.environ:
    print(
        f'r3frame {version.R3FRAME_YEAR}.{version.R3FRAME_MINOR}.{version.R3FRAME_PATCH} | "{random.choice(quotes.R3FRAME_QUOTES)}"'
    )


from r3frame.atom import Atom
from r3frame.globs import os, pg
from r3frame.util import abs_path
from r3frame.app.input import Mouse
from r3frame.util import point_inside

class ELEMENT_STATE:
    FILL: int = (1 << 0)
    HOVERED: int = (1 << 1)
    SHOW_TEXT: int = (1 << 2)
    SHOW_BORDER: int = (1 << 3)

# ------------------------------------------------------------ #
class Element(Atom):
    def __init__(
            self, text: str="Element",
            size: list[int]=[64, 64], color: list[int]=[0, 0, 0],
            padding: list[int]=[0, 0], location: list[int]=[0, 0],
            text_size: int=18, text_location: list[int]=[0, 0],
            text_color: list[int]=[255, 255, 255],
            font_path: str=abs_path("assets/fonts/megamax.ttf")
        ) -> None:
        super().__init__(0, 0)
        
        self.size = size
        self.location = location
        
        self.image_offset = [0, 0]
        self.image = pg.Surface(size)
        self.color: list[int] = color

        self.elements: dict[str, Element] = {}
        self.padding: list[int] = padding

        self.border_size = [1, 1]
        self.border_offset = [0, 0]
        self.border_radius = [0, 0, 0, 0]
        self.border_color = [255, 255, 255]

        self.text = text
        self.text_size: int = text_size
        self.text_color: list[int] = text_color
        self.text_location: list[int] = text_location

        self.font_path = font_path
        self.font: pg.Font = pg.Font(font_path, self.text_size)

        self.set_state(ELEMENT_STATE.FILL|ELEMENT_STATE.SHOW_TEXT)
    
    def load_font(self, font_path: str) -> None:
        if not isinstance(font_path, str) or not os.path.exists(font_path): return
        del self.font
        self.font_path = font_path
        self.font: pg.Font = pg.Font(font_path, self.text_size)

    def on_click(self) -> None: pass
    def on_hover(self) -> None: pass
    def on_unhover(self) -> None: pass

    def set_element(self, key: str, element: "Element") -> None:
        element.location = [
            element.location[0] + self.location[0],
            element.location[1] + self.location[1]
        ]
        self.elements[key] = element
    
    def get_element(self, key: str) -> "Element":
        return self.elements.get(key, None)
    
    def rem_element(self, key: str) -> None:
        if self.get_element(key) is not None:
            del self.elements[key]

    def border(self) -> pg.Rect:
        return pg.Rect(
            [(self.location[0] + self.border_offset[0]) - self.border_size[0],
             (self.location[1] + self.border_offset[1]) - self.border_size[1],], 
            [self.size[0] + self.border_size[0],
             self.size[1] + self.border_size[1],]
        )

    def render(self, surface: pg.Surface) -> None: 
        if self.get_state(ELEMENT_STATE.FILL):
            self.image.fill(self.color)
        
        if self.get_state(ELEMENT_STATE.SHOW_TEXT):
            self.image.blit(
                self.font.render(self.text, True, self.text_color, self.color),
                [self.text_location[0] + self.padding[0], self.text_location[1] + self.padding[1]]
            )

        surface.blit(self.image, [
            self.location[0] + self.image_offset[0],
            self.location[1] + self.image_offset[1]
        ])
        
        for element in self.elements.values():    # render elements
            element.render(surface)
        
        if self.get_state(ELEMENT_STATE.SHOW_BORDER):
            pg.draw.rect(
            surface=surface, color=self.border_color,
            rect=self.border(), width=self.border_size[0],
            border_top_left_radius=self.border_radius[0],
            border_top_right_radius=self.border_radius[1],
            border_bottom_left_radius=self.border_radius[2],
            border_bottom_right_radius=self.border_radius[3]
        )
            
    def update(self, event_manager) -> None:
        for element in self.elements.values():
            element.update(event_manager)
            mouse_within = point_inside(Mouse.location.screen, [
                element.location[0] - element.border_size[0], element.location[1] - element.border_size[1],
                element.size[0] + element.border_size[0], element.size[1] + element.border_size[1]
            ])
            if not element.get_state(ELEMENT_STATE.HOVERED) and mouse_within:
                Mouse.Hovering = Element
                element.set_state(ELEMENT_STATE.HOVERED)
                element.on_hover()
            if element.get_state(ELEMENT_STATE.HOVERED) and not mouse_within:
                Mouse.Hovering = None
                element.rem_state(ELEMENT_STATE.HOVERED)
                element.on_unhover()
            if element.get_state(ELEMENT_STATE.HOVERED) and event_manager.mouse_pressed(Mouse.LeftClick):
                event_manager.mouse[Mouse.LeftClick] = 0    # shouldnt need this but fixes the element double-click issue :|
                element.on_click()
# ------------------------------------------------------------ #
